@@ -10,25 +10,26 @@ import (
 	"time"
 
 	"github.com/argoproj-labs/applicationset/pkg/generators"
-	"github.com/argoproj/argo-cd/common"
-	"github.com/argoproj/argo-cd/pkg/apis/application/v1alpha1"
-	argov1alpha1 "github.com/argoproj/argo-cd/pkg/apis/application/v1alpha1"
-	"github.com/sirupsen/logrus"
+	"github.com/argoproj-labs/applicationset/pkg/utils"
+	argov1alpha1 "github.com/argoproj/argo-cd/v2/pkg/apis/application/v1alpha1"
 	log "github.com/sirupsen/logrus"
-	logtest "github.com/sirupsen/logrus/hooks/test"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/mock"
+	corev1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/tools/record"
-	"sigs.k8s.io/controller-runtime/pkg/client"
+	ctrl "sigs.k8s.io/controller-runtime"
 	crtclient "sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/client/fake"
 	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
 
 	argoprojiov1alpha1 "github.com/argoproj-labs/applicationset/api/v1alpha1"
-	appclientset "github.com/argoproj/argo-cd/pkg/client/clientset/versioned/fake"
-	dbmocks "github.com/argoproj/argo-cd/util/db/mocks"
+	appclientset "github.com/argoproj/argo-cd/v2/pkg/client/clientset/versioned/fake"
+	dbmocks "github.com/argoproj/argo-cd/v2/util/db/mocks"
+	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
+	kubefake "k8s.io/client-go/kubernetes/fake"
 )
 
 type generatorMock struct {
@@ -41,7 +42,7 @@ func (g *generatorMock) GetTemplate(appSetGenerator *argoprojiov1alpha1.Applicat
 	return args.Get(0).(*argoprojiov1alpha1.ApplicationSetTemplate)
 }
 
-func (g *generatorMock) GenerateParams(appSetGenerator *argoprojiov1alpha1.ApplicationSetGenerator) ([]map[string]string, error) {
+func (g *generatorMock) GenerateParams(appSetGenerator *argoprojiov1alpha1.ApplicationSetGenerator, _ *argoprojiov1alpha1.ApplicationSet) ([]map[string]string, error) {
 	args := g.Called(appSetGenerator)
 
 	return args.Get(0).([]map[string]string), args.Error(1)
@@ -57,7 +58,7 @@ func (g *generatorMock) GetRequeueAfter(appSetGenerator *argoprojiov1alpha1.Appl
 	return args.Get(0).(time.Duration)
 }
 
-func (r *rendererMock) RenderTemplateParams(tmpl *argov1alpha1.Application, params map[string]string) (*argov1alpha1.Application, error) {
+func (r *rendererMock) RenderTemplateParams(tmpl *argov1alpha1.Application, syncPolicy *argoprojiov1alpha1.ApplicationSetSyncPolicy, params map[string]string) (*argov1alpha1.Application, error) {
 	args := r.Called(tmpl, params)
 
 	if args.Error(1) != nil {
@@ -140,7 +141,7 @@ func TestExtractApplications(t *testing.T) {
 
 			rendererMock := rendererMock{}
 
-			expectedApps := []argov1alpha1.Application{}
+			var expectedApps []argov1alpha1.Application
 
 			if cc.generateParamsError == nil {
 				for _, p := range cc.params {
@@ -163,7 +164,8 @@ func TestExtractApplications(t *testing.T) {
 				Generators: map[string]generators.Generator{
 					"List": &generatorMock,
 				},
-				Renderer: &rendererMock,
+				Renderer:      &rendererMock,
+				KubeClientset: kubefake.NewSimpleClientset(),
 			}
 
 			got, err := r.generateApplications(argoprojiov1alpha1.ApplicationSet{
@@ -274,7 +276,8 @@ func TestMergeTemplateApplications(t *testing.T) {
 				Generators: map[string]generators.Generator{
 					"List": &generatorMock,
 				},
-				Renderer: &rendererMock,
+				Renderer:      &rendererMock,
+				KubeClientset: kubefake.NewSimpleClientset(),
 			}
 
 			got, _ := r.generateApplications(argoprojiov1alpha1.ApplicationSet{
@@ -736,11 +739,77 @@ func TestCreateOrUpdateInCluster(t *testing.T) {
 				},
 			},
 		},
+		{
+			name: "Ensure that argocd notifications state annotation is preserved from an existing app",
+			appSet: argoprojiov1alpha1.ApplicationSet{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "name",
+					Namespace: "namespace",
+				},
+				Spec: argoprojiov1alpha1.ApplicationSetSpec{
+					Template: argoprojiov1alpha1.ApplicationSetTemplate{
+						Spec: argov1alpha1.ApplicationSpec{
+							Project: "project",
+						},
+					},
+				},
+			},
+			existingApps: []argov1alpha1.Application{
+				{
+					TypeMeta: metav1.TypeMeta{
+						Kind:       "Application",
+						APIVersion: "argoproj.io/v1alpha1",
+					},
+					ObjectMeta: metav1.ObjectMeta{
+						Name:            "app1",
+						Namespace:       "namespace",
+						ResourceVersion: "2",
+						Labels:          map[string]string{"label-key": "label-value"},
+						Annotations: map[string]string{
+							"annot-key":           "annot-value",
+							NotifiedAnnotationKey: `{"b620d4600c771a6f4cxxxxxxx:on-deployed:[0].y7b5sbwa2Q329JYHxxxxxx-fBs:slack:slack-test":1617144614}`,
+						},
+					},
+					Spec: argov1alpha1.ApplicationSpec{
+						Project: "project",
+					},
+				},
+			},
+			desiredApps: []argov1alpha1.Application{
+				{
+					ObjectMeta: metav1.ObjectMeta{
+						Name: "app1",
+					},
+					Spec: argov1alpha1.ApplicationSpec{
+						Project: "project",
+					},
+				},
+			},
+			expected: []argov1alpha1.Application{
+				{
+					TypeMeta: metav1.TypeMeta{
+						Kind:       "Application",
+						APIVersion: "argoproj.io/v1alpha1",
+					},
+					ObjectMeta: metav1.ObjectMeta{
+						Name:            "app1",
+						Namespace:       "namespace",
+						ResourceVersion: "3",
+						Annotations: map[string]string{
+							NotifiedAnnotationKey: `{"b620d4600c771a6f4cxxxxxxx:on-deployed:[0].y7b5sbwa2Q329JYHxxxxxx-fBs:slack:slack-test":1617144614}`,
+						},
+					},
+					Spec: argov1alpha1.ApplicationSpec{
+						Project: "project",
+					},
+				},
+			},
+		},
 	} {
 
 		t.Run(c.name, func(t *testing.T) {
 
-			initObjs := []client.Object{&c.appSet}
+			initObjs := []crtclient.Object{&c.appSet}
 
 			for _, a := range c.existingApps {
 				err = controllerutil.SetControllerReference(&c.appSet, &a, scheme)
@@ -774,7 +843,7 @@ func TestCreateOrUpdateInCluster(t *testing.T) {
 	}
 }
 
-func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
+func TestRemoveFinalizerOnInvalidDestination_FinalizerTypes(t *testing.T) {
 
 	scheme := runtime.NewScheme()
 	err := argoprojiov1alpha1.AddToScheme(scheme)
@@ -783,18 +852,12 @@ func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
 	err = argov1alpha1.AddToScheme(scheme)
 	assert.Nil(t, err)
 
-	myCluster := argov1alpha1.Cluster{
-		Server: "https://kubernetes.default.svc",
-		Name:   "my-cluster2",
-	}
-
 	for _, c := range []struct {
 		// name is human-readable test name
 		name               string
 		existingFinalizers []string
 		expectedFinalizers []string
 	}{
-
 		{
 			name:               "no finalizers",
 			existingFinalizers: []string{},
@@ -802,10 +865,9 @@ func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
 		},
 		{
 			name:               "contains only argo finalizer",
-			existingFinalizers: []string{common.ResourcesFinalizerName},
+			existingFinalizers: []string{argov1alpha1.ResourcesFinalizerName},
 			expectedFinalizers: nil,
 		},
-
 		{
 			name:               "contains only non-argo finalizer",
 			existingFinalizers: []string{"non-argo-finalizer"},
@@ -813,7 +875,7 @@ func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
 		},
 		{
 			name:               "contains both argo and non-argo finalizer",
-			existingFinalizers: []string{"non-argo-finalizer", common.ResourcesFinalizerName},
+			existingFinalizers: []string{"non-argo-finalizer", argov1alpha1.ResourcesFinalizerName},
 			expectedFinalizers: []string{"non-argo-finalizer"},
 		},
 	} {
@@ -846,29 +908,44 @@ func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
 				},
 			}
 
-			initObjs := []client.Object{&app, &appSet}
+			initObjs := []crtclient.Object{&app, &appSet}
 
 			client := fake.NewClientBuilder().WithScheme(scheme).WithObjects(initObjs...).Build()
+			secret := &corev1.Secret{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "my-secret",
+					Namespace: "namespace",
+					Labels: map[string]string{
+						generators.ArgoCDSecretTypeLabel: generators.ArgoCDSecretTypeCluster,
+					},
+				},
+				Data: map[string][]byte{
+					// Since this test requires the cluster to be an invalid destination, we
+					// always return a cluster named 'my-cluster2' (different from app 'my-cluster', above)
+					"name":   []byte("mycluster2"),
+					"server": []byte("https://kubernetes.default.svc"),
+					"config": []byte("{\"username\":\"foo\",\"password\":\"foo\"}"),
+				},
+			}
 
-			// Always return a cluster named 'my-cluster2' (different from app 'my-cluster')
-			dbMock := &dbmocks.ArgoDB{}
-			dbMock.On("GetCluster", mock.Anything, "https://kubernetes.default.svc").Return(&myCluster, nil)
-			dbMock.On("ListClusters", mock.Anything).Return(&v1alpha1.ClusterList{Items: []argov1alpha1.Cluster{
-				myCluster,
-			}}, nil)
+			objects := append([]runtime.Object{}, secret)
+			kubeclientset := kubefake.NewSimpleClientset(objects...)
 
 			r := ApplicationSetReconciler{
-				Client:   client,
-				Scheme:   scheme,
-				Recorder: record.NewFakeRecorder(10),
-				ArgoDB:   dbMock,
+				Client:        client,
+				Scheme:        scheme,
+				Recorder:      record.NewFakeRecorder(10),
+				KubeClientset: kubeclientset,
 			}
 
+			clusterList, err := utils.ListClusters(context.Background(), kubeclientset, "namespace")
+			assert.NoError(t, err, "Unexpected error")
+
 			appLog := log.WithFields(log.Fields{"app": app.Name, "appSet": ""})
 
 			appInputParam := app.DeepCopy()
 
-			err = r.removeFinalizerOnInvalidDestination(context.Background(), appSet, appInputParam, appLog)
+			err = r.removeFinalizerOnInvalidDestination(context.Background(), appSet, appInputParam, clusterList, appLog)
 			assert.NoError(t, err, "Unexpected error")
 
 			retrievedApp := argov1alpha1.Application{}
@@ -888,6 +965,163 @@ func TestRemoveFinalizerOnInvalidDestination(t *testing.T) {
 	}
 }
 
+func TestRemoveFinalizerOnInvalidDestination_DestinationTypes(t *testing.T) {
+
+	scheme := runtime.NewScheme()
+	err := argoprojiov1alpha1.AddToScheme(scheme)
+	assert.Nil(t, err)
+
+	err = argov1alpha1.AddToScheme(scheme)
+	assert.Nil(t, err)
+
+	for _, c := range []struct {
+		// name is human-readable test name
+		name                   string
+		destinationField       argov1alpha1.ApplicationDestination
+		expectFinalizerRemoved bool
+	}{
+		{
+			name: "invalid cluster: empty destination",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+			},
+			expectFinalizerRemoved: true,
+		},
+		{
+			name: "invalid cluster: invalid server url",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Server:    "https://1.2.3.4",
+			},
+			expectFinalizerRemoved: true,
+		},
+		{
+			name: "invalid cluster: invalid cluster name",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Name:      "invalid-cluster",
+			},
+			expectFinalizerRemoved: true,
+		},
+		{
+			name: "invalid cluster by both valid",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Name:      "mycluster2",
+				Server:    "https://kubernetes.default.svc",
+			},
+			expectFinalizerRemoved: true,
+		},
+		{
+			name: "invalid cluster by both invalid",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Name:      "mycluster3",
+				Server:    "https://4.5.6.7",
+			},
+			expectFinalizerRemoved: true,
+		},
+		{
+			name: "valid cluster by name",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Name:      "mycluster2",
+			},
+			expectFinalizerRemoved: false,
+		},
+		{
+			name: "valid cluster by server",
+			destinationField: argov1alpha1.ApplicationDestination{
+				Namespace: "namespace",
+				Server:    "https://kubernetes.default.svc",
+			},
+			expectFinalizerRemoved: false,
+		},
+	} {
+
+		t.Run(c.name, func(t *testing.T) {
+
+			appSet := argoprojiov1alpha1.ApplicationSet{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "name",
+					Namespace: "namespace",
+				},
+				Spec: argoprojiov1alpha1.ApplicationSetSpec{
+					Template: argoprojiov1alpha1.ApplicationSetTemplate{
+						Spec: argov1alpha1.ApplicationSpec{
+							Project: "project",
+						},
+					},
+				},
+			}
+
+			app := argov1alpha1.Application{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:       "app1",
+					Finalizers: []string{argov1alpha1.ResourcesFinalizerName},
+				},
+				Spec: argov1alpha1.ApplicationSpec{
+					Project:     "project",
+					Source:      argov1alpha1.ApplicationSource{Path: "path", TargetRevision: "revision", RepoURL: "repoURL"},
+					Destination: c.destinationField,
+				},
+			}
+
+			initObjs := []crtclient.Object{&app, &appSet}
+
+			client := fake.NewClientBuilder().WithScheme(scheme).WithObjects(initObjs...).Build()
+			secret := &corev1.Secret{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "my-secret",
+					Namespace: "namespace",
+					Labels: map[string]string{
+						generators.ArgoCDSecretTypeLabel: generators.ArgoCDSecretTypeCluster,
+					},
+				},
+				Data: map[string][]byte{
+					// Since this test requires the cluster to be an invalid destination, we
+					// always return a cluster named 'my-cluster2' (different from app 'my-cluster', above)
+					"name":   []byte("mycluster2"),
+					"server": []byte("https://kubernetes.default.svc"),
+					"config": []byte("{\"username\":\"foo\",\"password\":\"foo\"}"),
+				},
+			}
+
+			objects := append([]runtime.Object{}, secret)
+			kubeclientset := kubefake.NewSimpleClientset(objects...)
+
+			r := ApplicationSetReconciler{
+				Client:        client,
+				Scheme:        scheme,
+				Recorder:      record.NewFakeRecorder(10),
+				KubeClientset: kubeclientset,
+			}
+
+			clusterList, err := utils.ListClusters(context.Background(), kubeclientset, "namespace")
+			assert.NoError(t, err, "Unexpected error")
+
+			appLog := log.WithFields(log.Fields{"app": app.Name, "appSet": ""})
+
+			appInputParam := app.DeepCopy()
+
+			err = r.removeFinalizerOnInvalidDestination(context.Background(), appSet, appInputParam, clusterList, appLog)
+			assert.NoError(t, err, "Unexpected error")
+
+			retrievedApp := argov1alpha1.Application{}
+			err = client.Get(context.Background(), crtclient.ObjectKeyFromObject(&app), &retrievedApp)
+			assert.NoError(t, err, "Unexpected error")
+
+			finalizerRemoved := len(retrievedApp.Finalizers) == 0
+
+			assert.True(t, c.expectFinalizerRemoved == finalizerRemoved)
+
+			bytes, _ := json.MarshalIndent(retrievedApp, "", "  ")
+			t.Log("Contents of app after call:", string(bytes))
+
+		})
+	}
+}
+
 func TestCreateApplications(t *testing.T) {
 
 	scheme := runtime.NewScheme()
@@ -1047,7 +1281,7 @@ func TestCreateApplications(t *testing.T) {
 			},
 		},
 	} {
-		initObjs := []client.Object{&c.appSet}
+		initObjs := []crtclient.Object{&c.appSet}
 		for _, a := range c.existsApps {
 			err = controllerutil.SetControllerReference(&c.appSet, &a, scheme)
 			assert.Nil(t, err)
@@ -1189,7 +1423,7 @@ func TestDeleteInCluster(t *testing.T) {
 			},
 		},
 	} {
-		initObjs := []client.Object{&c.appSet}
+		initObjs := []crtclient.Object{&c.appSet}
 		for _, a := range c.existingApps {
 			temp := a
 			err = controllerutil.SetControllerReference(&c.appSet, &temp, scheme)
@@ -1200,9 +1434,10 @@ func TestDeleteInCluster(t *testing.T) {
 		client := fake.NewClientBuilder().WithScheme(scheme).WithObjects(initObjs...).Build()
 
 		r := ApplicationSetReconciler{
-			Client:   client,
-			Scheme:   scheme,
-			Recorder: record.NewFakeRecorder(len(initObjs) + len(c.expected)),
+			Client:        client,
+			Scheme:        scheme,
+			Recorder:      record.NewFakeRecorder(len(initObjs) + len(c.expected)),
+			KubeClientset: kubefake.NewSimpleClientset(),
 		}
 
 		err = r.deleteInCluster(context.TODO(), c.appSet, c.desiredApps)
@@ -1281,382 +1516,6 @@ func TestGetMinRequeueAfter(t *testing.T) {
 
 	assert.Equal(t, time.Duration(1)*time.Second, got)
 }
-
-func TestInvalidGenerators(t *testing.T) {
-
-	scheme := runtime.NewScheme()
-	err := argoprojiov1alpha1.AddToScheme(scheme)
-	assert.Nil(t, err)
-	err = argov1alpha1.AddToScheme(scheme)
-	assert.Nil(t, err)
-
-	for _, c := range []struct {
-		testName        string
-		appSet          argoprojiov1alpha1.ApplicationSet
-		expectedInvalid bool
-		expectedNames   map[string]bool
-	}{
-		{
-			testName: "valid generators, with annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-								"generators":[
-									{"list":{}},
-									{"cluster":{}},
-									{"git":{}}
-								]
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     &argoprojiov1alpha1.ListGenerator{},
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: &argoprojiov1alpha1.ClusterGenerator{},
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      &argoprojiov1alpha1.GitGenerator{},
-						},
-					},
-				},
-			},
-			expectedInvalid: false,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "invalid generators, no annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "valid and invalid generators, no annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: &argoprojiov1alpha1.ClusterGenerator{},
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      &argoprojiov1alpha1.GitGenerator{},
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "valid and invalid generators, with annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-								"generators":[
-									{"cluster":{}},
-									{"bbb":{}},
-									{"git":{}},
-									{"aaa":{}}
-								]
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: &argoprojiov1alpha1.ClusterGenerator{},
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      &argoprojiov1alpha1.GitGenerator{},
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames: map[string]bool{
-				"aaa": true,
-				"bbb": true,
-			},
-		},
-		{
-			testName: "invalid generator, annotation with missing spec",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "invalid generator, annotation with missing generators array",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "invalid generator, annotation with empty generators array",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-								"generators":[
-								]
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-		{
-			testName: "invalid generator, annotation with empty generator",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "name",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-								"generators":[
-									{}
-								]
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedInvalid: true,
-			expectedNames:   map[string]bool{},
-		},
-	} {
-		hasInvalid, names := invalidGenerators(&c.appSet)
-		assert.Equal(t, c.expectedInvalid, hasInvalid, c.testName)
-		assert.Equal(t, c.expectedNames, names, c.testName)
-	}
-}
-
-func TestCheckInvalidGenerators(t *testing.T) {
-
-	scheme := runtime.NewScheme()
-	err := argoprojiov1alpha1.AddToScheme(scheme)
-	assert.Nil(t, err)
-	err = argov1alpha1.AddToScheme(scheme)
-	assert.Nil(t, err)
-
-	for _, c := range []struct {
-		testName    string
-		appSet      argoprojiov1alpha1.ApplicationSet
-		expectedMsg string
-	}{
-		{
-			testName: "invalid generator, without annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "test-app-set",
-					Namespace: "namespace",
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     &argoprojiov1alpha1.ListGenerator{},
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      &argoprojiov1alpha1.GitGenerator{},
-						},
-					},
-				},
-			},
-			expectedMsg: "ApplicationSet test-app-set contains unrecognized generators",
-		},
-		{
-			testName: "invalid generator, with annotation",
-			appSet: argoprojiov1alpha1.ApplicationSet{
-				ObjectMeta: metav1.ObjectMeta{
-					Name:      "test-app-set",
-					Namespace: "namespace",
-					Annotations: map[string]string{
-						"kubectl.kubernetes.io/last-applied-configuration": `{
-							"spec":{
-								"generators":[
-									{"list":{}},
-									{"bbb":{}},
-									{"git":{}},
-									{"aaa":{}}
-								]
-							}
-						}`,
-					},
-				},
-				Spec: argoprojiov1alpha1.ApplicationSetSpec{
-					Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
-						{
-							List:     &argoprojiov1alpha1.ListGenerator{},
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      &argoprojiov1alpha1.GitGenerator{},
-						},
-						{
-							List:     nil,
-							Clusters: nil,
-							Git:      nil,
-						},
-					},
-				},
-			},
-			expectedMsg: "ApplicationSet test-app-set contains unrecognized generators: aaa, bbb",
-		},
-	} {
-		oldhooks := logrus.StandardLogger().ReplaceHooks(logrus.LevelHooks{})
-		defer logrus.StandardLogger().ReplaceHooks(oldhooks)
-		hook := logtest.NewGlobal()
-
-		checkInvalidGenerators(&c.appSet)
-		assert.True(t, len(hook.Entries) >= 1, c.testName)
-		assert.Equal(t, logrus.WarnLevel, hook.LastEntry().Level, c.testName)
-		assert.Equal(t, c.expectedMsg, hook.LastEntry().Message, c.testName)
-		hook.Reset()
-	}
-}
-
 func TestHasDuplicateNames(t *testing.T) {
 
 	scheme := runtime.NewScheme()
@@ -1875,9 +1734,27 @@ func TestValidateGeneratedApplications(t *testing.T) {
 
 		t.Run(cc.name, func(t *testing.T) {
 
+			secret := &corev1.Secret{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "my-secret",
+					Namespace: "namespace",
+					Labels: map[string]string{
+						generators.ArgoCDSecretTypeLabel: generators.ArgoCDSecretTypeCluster,
+					},
+				},
+				Data: map[string][]byte{
+					"name":   []byte("my-cluster"),
+					"server": []byte("https://kubernetes.default.svc"),
+					"config": []byte("{\"username\":\"foo\",\"password\":\"foo\"}"),
+				},
+			}
+
+			objects := append([]runtime.Object{}, secret)
+			kubeclientset := kubefake.NewSimpleClientset(objects...)
+
 			argoDBMock := dbmocks.ArgoDB{}
 			argoDBMock.On("GetCluster", mock.Anything, "https://kubernetes.default.svc").Return(&myCluster, nil)
-			argoDBMock.On("ListClusters", mock.Anything).Return(&v1alpha1.ClusterList{Items: []argov1alpha1.Cluster{
+			argoDBMock.On("ListClusters", mock.Anything).Return(&argov1alpha1.ClusterList{Items: []argov1alpha1.Cluster{
 				myCluster,
 			}}, nil)
 
@@ -1893,6 +1770,7 @@ func TestValidateGeneratedApplications(t *testing.T) {
 				Generators:       map[string]generators.Generator{},
 				ArgoDB:           &argoDBMock,
 				ArgoAppClientset: appclientset.NewSimpleClientset(argoObjs...),
+				KubeClientset:    kubeclientset,
 			}
 
 			appSetInfo := argoprojiov1alpha1.ApplicationSet{}
@@ -1914,3 +1792,61 @@ func TestValidateGeneratedApplications(t *testing.T) {
 		})
 	}
 }
+
+func TestReconcilerValidationErrorBehaviour(t *testing.T) {
+
+	scheme := runtime.NewScheme()
+	err := argoprojiov1alpha1.AddToScheme(scheme)
+	assert.Nil(t, err)
+	err = argov1alpha1.AddToScheme(scheme)
+	assert.Nil(t, err)
+
+	appSet := argoprojiov1alpha1.ApplicationSet{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "name",
+			Namespace: "argocd",
+		},
+		Spec: argoprojiov1alpha1.ApplicationSetSpec{
+			Generators: []argoprojiov1alpha1.ApplicationSetGenerator{
+				{List: &argoprojiov1alpha1.ListGenerator{
+					Elements: []apiextensionsv1.JSON{{
+						Raw: []byte(`{"cluster": "my-cluster","url": "https://kubernetes.default.svc"}`),
+					}},
+				}},
+			},
+			Template: argoprojiov1alpha1.ApplicationSetTemplate{},
+		},
+	}
+	kubeclientset := kubefake.NewSimpleClientset([]runtime.Object{}...)
+	argoDBMock := dbmocks.ArgoDB{}
+	argoObjs := []runtime.Object{}
+
+	client := fake.NewClientBuilder().WithScheme(scheme).WithObjects(&appSet).Build()
+
+	r := ApplicationSetReconciler{
+		Log:      ctrl.Log.WithName("controllers").WithName("ApplicationSet"),
+		Client:   client,
+		Scheme:   scheme,
+		Renderer: &utils.Render{},
+		Recorder: record.NewFakeRecorder(1),
+		Generators: map[string]generators.Generator{
+			"List": generators.NewListGenerator(),
+		},
+		ArgoDB:           &argoDBMock,
+		ArgoAppClientset: appclientset.NewSimpleClientset(argoObjs...),
+		KubeClientset:    kubeclientset,
+	}
+
+	req := ctrl.Request{
+		NamespacedName: types.NamespacedName{
+			Namespace: "argocd",
+			Name:      "name",
+		},
+	}
+
+	// Verify that on validation error, no error is returned, but the object is requeued
+	res, err := r.Reconcile(context.Background(), req)
+	assert.Nil(t, err)
+	assert.True(t, res.RequeueAfter > 0)
+
+}
